#include<string.h>
#include<stdlib.h>
#include"stree.h"
#include"file_hdl.h"
#include"tree_hdl.h"

union stackdata {
  long filep;
  void *memp;
};

struct mystack {
  int count;
  int reserved;
  union stackdata *data;
};

/* Sort the stack so that smallest keys are on top.
   Actually, you souldn't be able to do this to a stack. 
*/
int cmpsdata(const void *one,const void *two)
{
  return ((union stackdata*)two)->filep - ((union stackdata*)one)->filep;
}

void sort_stack(struct mystack *this)
{
  if(this->count>1) {
    qsort(this->data,this->count,sizeof(union stackdata),cmpsdata);
  }
}

void init_stack(struct mystack *here)
{
  here->count=0;
  here->reserved=10;
  here->data=malloc(sizeof(union stackdata)*here->reserved);
  if(here->data==NULL) {
    mem_error();
  }
}

void adjust_size(struct mystack *here)
{
  if(here->count>=here->reserved) {
    here->reserved*=2;
    here->data=realloc(here->data,sizeof(union stackdata)*here->reserved);
    if(here->data==NULL) {
      mem_error();
    }
  } else if(here->count<(here->reserved>>1)) {
    here->reserved/=2;
    here->data=realloc(here->data,sizeof(union stackdata)*here->reserved);
    if(here->data==NULL) {
      mem_error();
    }
  }
}
void push_long(struct mystack *here,long this)
{
  adjust_size(here);
  here->data[here->count++].filep=this;
}

void push_ptr(struct mystack *here,void *pointer)
{
  adjust_size(here);
  here->data[here->count++].memp=pointer;
}


int empty_stack(struct mystack *here) {
  return !here->count;
}

void free_stack(struct mystack *here) {
  here->count=here->reserved=0;
  free(here->data);
}

union stackdata pop(struct mystack *here)
{
  adjust_size(here);
  if(here->count>0) {
    return here->data[--here->count];
  };
  /* else {
    return ;
  }*/
}

struct mystack give_hits(st_meta *data)
{
  int i,pstart=0,plen=strlen((char*)data->pattern);
  st_trans *look=NULL;
  st_node *my=&data->root;
  struct mystack result, traverse;
  union stackdata tmp;
  char *compar;

  init_stack(&result);


  for(i=0;i<plen;) {
    valid_transitions(data,my);
    look=get_transition(data,my,data->pattern[i]);

    if(look==NULL) {
      return result;
    }
    pstart=look->begin-i;
    if(look->end>=0) {
      i=i+look->end-look->begin+1;
    } else {
      i=plen;
    }

    my=look->childs;
  }

  /* We need only one access to the text. */
  compar=malloc(plen);
  if(compar==NULL) {
    mem_error();
  }
  read_string(data,compar,pstart,pstart+plen-1,plen-1);
  if(strncmp(compar,(char*)data->pattern,strlen(compar))!=0) {
    return result;
  }
  free(compar);


  /* the pattern does exist in the text. Let's find out where. 
     i.e. traverse subtree from my. */

  /* This is a bit tree-implementation specific. */

  init_stack(&traverse);

  push_ptr(&traverse,my);

  while(!empty_stack(&traverse) && my!=NULL) {
    tmp=pop(&traverse);
    my=(st_node*)tmp.memp;
    valid_transitions(data,my);
    for(i=0;i<my->n_tran;i++) {
      if(my->trans[i].end>=0) {
	push_ptr(&traverse,my->trans[i].childs);
      } else {
	push_long(&result,-my->trans[i].end-1);
      }
    }
  }
  /* Special case: Only one hit.*/
  if(my==NULL) {
    push_long(&result,-look->end-1);
  }

  free_stack(&traverse);

  return result;
}

void stree_find(st_meta *data)
{
  int plen;
  long begin,end;
  char *tmpstr;
  struct mystack hits;
  union stackdata tmp;

  hits=give_hits(data);


  if(empty_stack(&hits)) {
    printf("Could not find \"%s\"\n",data->pattern);
  }

  sort_stack(&hits);
  plen=strlen((char*)data->pattern);
  tmpstr=(char*)malloc(plen*3+1);
  if(tmpstr==NULL) {
    mem_error();
  }

  while(!empty_stack(&hits)) {
    tmp=pop(&hits);

    begin=tmp.filep-plen;
    if(begin<0) {
      begin=0;
    }
    end=tmp.filep+2*plen-1;
    if(end>data->fsize) {
      end=data->fsize;
    }
    read_string(data,tmpstr,begin,end,plen*3);
    printf("offset %ld: ...%s...\n",tmp.filep,tmpstr);
  }
  free_stack(&hits);
  free(tmpstr);
}

