#ifndef _STREE_H_
#define _STREE_H_

#include<time.h>
#include<sys/stat.h>
#include<unistd.h>
#include<limits.h>
#include<stdio.h>

/* Create suffix tree using Esko Ukkonen's algorithm. O(T) */


typedef unsigned char byte;
typedef unsigned int idx_p;

/* Suffix tree transition data. */
typedef struct stree_transit {
  /* First letter of the transition. */
  byte transId;
  /* Link to the node. */
  struct stree_node *childs;

  /* Pointers to the textfile that represent this transition. */
  /* Beginning of this transition. */
  signed long begin;

  /* if end>=begin, end of this transition.
     else end=-bgs-1  bgs is beginning of suffix represented by this leaf.
           (bgs=-end-1)
  */
  signed long end;
} st_trans;

#define INFTY_P LONG_MAX

/* Suffix tree node. */
typedef struct stree_node {
  /* Number of transitions from this node. */
  byte n_tran;
  /* Transitions in ascending order. */
  struct stree_transit *trans;
  /* If trans==NULL then transitions are on disk_trans on disk 
  */
  idx_p disk_trans;

  /* Suffix link. */
  struct stree_node *sfx_link;

} st_node;


/* Suffix tree MetaData to be stored in RAM */

struct txtcache {
  unsigned long lowstart;
  char *lowdata;
  unsigned long highstart;
  char *highdata;
  unsigned long highlimit;
  /* Statistics */
  unsigned long uses;
  unsigned long highmiss;
  unsigned long lowmiss;
};


typedef struct stree_meta {
  /* Size of the indexed file */
  size_t fsize;
  /* Time of modification of the indexed file */
  time_t fmodified;
  /* Name of the indexed file. */
  char fname[256];
  /* Usage mode: c for create, f for find. */
  char use_mode;
  /* Pattern to find. */
  byte *pattern;
  /* Size of text_p -pointer, in bytes (1-4) (disregarded) */
  byte text_p_size;
  /* 0 if suffix links are not stored. (for now allways 0) */
  byte with_sfx_links;
  /* The root node. */
  struct stree_node root;
  /* The text-file descriptor. */
  int txt_filedes;
  /* Descriptor of the indexfile.
   size of the index < 28*size of text  i.e. O(T) */
  int idx_filedes;
  struct txtcache cache;
} st_meta;


void create_stree(st_meta *data);
void mem_error();
void stree_find(st_meta *data);

#endif
