#include<stdlib.h>
#include<stdio.h>
#include<string.h>

#include "stree.h"
#include "tree_hdl.h"
#include "file_hdl.h"


static st_trans top_trans;

/* Just a common binarysearch. */
st_trans *get_transition(st_meta *sm, st_node *here, byte t)
{
  int bgin=0,end,mid;


  if(here==NULL) {
    /* From NULL (= |_ ) are transitions to everywhere. */
    top_trans.transId=t;
    top_trans.begin=top_trans.end=(long)-t;
    top_trans.childs=&(sm->root);

    return &top_trans;
  }

  end=here->n_tran-1;

  valid_transitions(sm,here);
  if(here->trans==NULL)
    return NULL;

  mid=(bgin+end)/2;

  /* The array is in ASCENDING order. (Lagest in the end) */
  while(bgin<end) {
    if(t>here->trans[mid].transId) {
      bgin=mid+1;
    } else if(t<here->trans[mid].transId) {
      end=mid-1;
    } else {
      return &(here->trans[mid]);
    }
    mid=(bgin+end)/2;
  }
  if(t==here->trans[mid].transId) {
    return &(here->trans[mid]);
  } else {
    return NULL;
  }
}

/* Let's keep the transition array in ascending order. */
st_trans *insert_transition(st_meta *sm, st_node *here, st_trans *this)
{
  int i;

  here->n_tran++;

  here->trans=realloc(here->trans,sizeof(st_trans)*here->n_tran);
  if(here->trans==NULL) {
    mem_error();
  }

  /* We'll do an insertion sort kind of thing here. */
  for(i=here->n_tran-1;i>0 && this->transId<here->trans[i-1].transId;i--) {
    here->trans[i]=here->trans[i-1];
    /*here->trans[i]=tmp;*/
  }
  /* Special case. */
  if(i<0) i=0;

  memcpy(&here->trans[i],this,sizeof(st_trans));
  return &(here->trans[i]);
}

void print_trans(st_meta *sm, st_trans *trans,int spaces)
{
  int bgin,end,len,i;
  char *str;

  bgin=trans->begin;

  if(trans->end==INFTY_P|| trans->end<0) {
    if(sm->fsize<(bgin+21)) {
      end=sm->fsize-1;
    } else {
      end=bgin+20;
    }
  } else {
    end=trans->end;
  }

  len=spaces+end-bgin+2;

  str=(char*)malloc(len);
  if(str==NULL) {
    mem_error();
  }

  memset(str,'\t',spaces);
  read_string(sm,str+spaces,bgin,end,20);

  for(i=0;i<strlen(str);i++) {
    if(str[i]=='\n')
      str[i]='\\';
  }

  if(end!=trans->end)
    printf("%s... (bgn=%ld)\n",str,-trans->end-1);
  else
    printf("%s\n",str);

  free(str);

}

void print_nodes(st_meta *sm, st_node *node,int spaces)
{
  int i;

  valid_transitions(sm,node);

  for(i=0;i<node->n_tran;i++) {
    print_trans(sm,&node->trans[i],spaces);
    if(node->trans[i].childs!=NULL) {
      print_nodes(sm,node->trans[i].childs,spaces+1);
    }
  }
}

void print_tree(st_meta *sm)
{
  print_nodes(sm,&sm->root,0);
}

